@@ -96,8 +96,8 @@ NTSTATUS
 // Worker thread count
 //
 
-#define AFS_WORKER_COUNT        5
-#define AFS_IO_WORKER_COUNT     5
+#define AFS_WORKER_COUNT        16
+#define AFS_IO_WORKER_COUNT     8
 
 //
 // Worker thread states
@@ -434,6 +434,10 @@ DEFINE_GUID (GUID_SD_AFS_REDIRECTOR_CONTROL_OBJECT,
 #define IO_REPARSE_TAG_OPENAFS_DFS              0x00000037L
 #endif
 
+#ifndef IO_REPARSE_TAG_SURROGATE
+#define IO_REPARSE_TAG_SURROGATE                0x20000000L
+#endif
+
 //  {EF21A155-5C92-4470-AB3B-370403D96369}
 DEFINE_GUID (GUID_AFS_REPARSE_GUID,
         0xEF21A155, 0x5C92, 0x4470, 0xAB, 0x3B, 0x37, 0x04, 0x03, 0xD9, 0x63, 0x69);
